# Copyright (c) OpenMMLab. All rights reserved.
"""This module provides unified file I/O related functions, which support
operating I/O with different file backends based on the specified filepath or
backend_args.

MMEngine currently supports five file backends:

- LocalBackend
- Boto3Backend
- HTTPBackend
- LmdbBackend
- MemcacheBackend

Note that this module provide a union of all of the above file backends so
NotImplementedError will be raised if the interface in the file backend is not
implemented.

There are two ways to call a method of a file backend:

- Initialize a file backend with ``get_file_backend`` and call its methods.
- Directory call unified I/O functions, which will call ``get_file_backend``
  first and then call the corresponding backend method.

Examples:
    >>> # Initialize a file backend and call its methods
    >>> import mmengine.fileio as fileio
    >>> backend = fileio.get_file_backend(backend_args={'backend': 's3'})
    >>> backend.get('s3://path/of/your/file')

    >>> # Directory call unified I/O functions
    >>> fileio.get('s3://path/of/your/file')
"""
import json
import warnings
from contextlib import contextmanager
from io import BytesIO, StringIO
from pathlib import Path
from typing import Generator, Iterator, Optional, Tuple, Union

from jammy.io.path import is_filepath

from .backends import backends, prefix_to_backends
from .file_client import FileClient

# file_handlers and register_handler had been moved to
# mmengine/fileio/handlers/registry_utis. Import them
# in this file to keep backward compatibility.
from .handlers import file_handlers, register_handler  # noqa: F401

backend_instances: dict = {}


def _parse_uri_prefix(uri: Union[str, Path]) -> str:
    """Parse the prefix of uri.

    Args:
        uri (str or Path): Uri to be parsed that contains the file prefix.

    Examples:
        >>> _parse_uri_prefix('/home/path/of/your/file')
        ''
        >>> _parse_uri_prefix('s3://path/of/your/file')
        's3'
        >>> _parse_uri_prefix('clusterName:s3://path/of/your/file')
        's3'

    Returns:
        str: Return the prefix of uri if the uri contains '://'. Otherwise,
        return ''.
    """
    assert is_filepath(uri)
    uri = str(uri)
    # if uri does not contains '://', the uri will be handled by
    # LocalBackend by default
    if "://" not in uri:
        return ""
    else:
        prefix, _ = uri.split("://")
        # In the case of Boto3Backend, the prefix may contain the cluster
        # name like clusterName:s3://path/of/your/file
        if ":" in prefix:
            _, prefix = prefix.split(":")
        return prefix


def _get_file_backend(prefix: str, backend_args: dict):
    """Return a file backend based on the prefix or backend_args.

    Args:
        prefix (str): Prefix of uri.
        backend_args (dict): Arguments to instantiate the corresponding
            backend.
    """
    # backend name has a higher priority
    if "backend" in backend_args:
        # backend_args should not be modified
        backend_args_bak = backend_args.copy()
        backend_name = backend_args_bak.pop("backend")
        backend = backends[backend_name](**backend_args_bak)
    else:
        backend = prefix_to_backends[prefix](**backend_args)
    return backend


def set_s3_backend(
    key: str = "s3:{}",
    backend_args: Optional[dict] = None,
):
    """register s3 backend.

    Args:
        key str: The key to register the s3 backend. Defaults to s3.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
    """
    global backend_instances
    if backend_args is None:
        backend_args = {}
    backend = _get_file_backend(key, backend_args)
    backend_instances[key] = backend
    return backend


def get_file_backend(
    uri: Union[str, Path, None] = None,
    *,
    backend_args: Optional[dict] = None,
    enable_singleton: bool = False,
    backend_key: Optional[str] = None,
):
    """Return a file backend based on the prefix of uri or backend_args.

    Args:
        uri (str or Path): Uri to be parsed that contains the file prefix.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        enable_singleton (bool): Whether to enable the singleton pattern.
            If it is True, the backend created will be reused if the
            signature is same with the previous one. Defaults to False.
        backend_key: str: The key to register the backend. Defaults to None.

    Returns:
        BaseStorageBackend: Instantiated Backend object.

    Examples:
        >>> # get file backend based on the prefix of uri
        >>> uri = 's3://path/of/your/file'
        >>> backend = get_file_backend(uri)
        >>> # get file backend based on the backend_args
        >>> backend = get_file_backend(backend_args={'backend': 's3'})
        >>> # backend name has a higher priority if 'backend' in backend_args
        >>> backend = get_file_backend(uri, backend_args={'backend': 's3'})
    """
    global backend_instances
    if backend_key is not None:
        if backend_key in backend_instances:
            return backend_instances[backend_key]

    if backend_args is None:
        backend_args = {}

    if uri is None and "backend" not in backend_args and backend_key is None:
        raise ValueError(
            'uri should not be None when "backend" does not exist in '
            "backend_args and backend_key is None"
        )

    if uri is not None:
        prefix = _parse_uri_prefix(uri)
    else:
        prefix = ""

    if enable_singleton:
        # TODO: whether to pass sort_key to json.dumps
        unique_key = f"{prefix}:{json.dumps(backend_args)}"
        if unique_key in backend_instances:
            return backend_instances[unique_key]

        backend = _get_file_backend(prefix, backend_args)
        backend_instances[unique_key] = backend
        if backend_key is not None:
            backend_instances[backend_key] = backend
        return backend
    else:
        backend = _get_file_backend(prefix, backend_args)
        return backend


def get(
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> bytes:
    """Read bytes from a given ``filepath`` with 'rb' mode.

    Args:
        filepath (str or Path): Path to read data.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        bytes: Expected bytes object.

    Examples:
        >>> filepath = '/path/of/file'
        >>> get(filepath)
        b'hello world'
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    return backend.get(filepath)


def get_text(
    filepath: Union[str, Path],
    encoding="utf-8",
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> str:
    """Read text from a given ``filepath`` with 'r' mode.

    Args:
        filepath (str or Path): Path to read data.
        encoding (str): The encoding format used to open the ``filepath``.
            Defaults to 'utf-8'.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        str: Expected text reading from ``filepath``.

    Examples:
        >>> filepath = '/path/of/file'
        >>> get_text(filepath)
        'hello world'
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    return backend.get_text(filepath, encoding)


def put(
    obj: bytes,
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> None:
    """Write bytes to a given ``filepath`` with 'wb' mode.

    Note:
        ``put`` should create a directory if the directory of
        ``filepath`` does not exist.

    Args:
        obj (bytes): Data to be written.
        filepath (str or Path): Path to write data.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Examples:
        >>> filepath = '/path/of/file'
        >>> put(b'hello world', filepath)
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    backend.put(obj, filepath)


def put_text(
    obj: str,
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> None:
    """Write text to a given ``filepath`` with 'w' mode.

    Note:
        ``put_text`` should create a directory if the directory of
        ``filepath`` does not exist.

    Args:
        obj (str): Data to be written.
        filepath (str or Path): Path to write data.
        encoding (str, optional): The encoding format used to open the
            ``filepath``. Defaults to 'utf-8'.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Examples:
        >>> filepath = '/path/of/file'
        >>> put_text('hello world', filepath)
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    backend.put_text(obj, filepath)


def exists(
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> bool:
    """Check whether a file path exists.

    Args:
        filepath (str or Path): Path to be checked whether exists.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        bool: Return ``True`` if ``filepath`` exists, ``False`` otherwise.

    Examples:
        >>> filepath = '/path/of/file'
        >>> exists(filepath)
        True
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    return backend.exists(filepath)


def isdir(
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> bool:
    """Check whether a file path is a directory.

    Args:
        filepath (str or Path): Path to be checked whether it is a
            directory.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        bool: Return ``True`` if ``filepath`` points to a directory,
        ``False`` otherwise.

    Examples:
        >>> filepath = '/path/of/dir'
        >>> isdir(filepath)
        True
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    return backend.isdir(filepath)


def isfile(
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> bool:
    """Check whether a file path is a file.

    Args:
        filepath (str or Path): Path to be checked whether it is a file.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        bool: Return ``True`` if ``filepath`` points to a file, ``False``
        otherwise.

    Examples:
        >>> filepath = '/path/of/file'
        >>> isfile(filepath)
        True
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    return backend.isfile(filepath)


def join_path(
    filepath: Union[str, Path],
    *filepaths: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    r"""Concatenate all file paths.

    Join one or more filepath components intelligently. The return value
    is the concatenation of filepath and any members of \*filepaths.

    Args:
        filepath (str or Path): Path to be concatenated.
        *filepaths (str or Path): Other paths to be concatenated.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        str: The result of concatenation.

    Examples:
        >>> filepath1 = '/path/of/dir1'
        >>> filepath2 = 'dir2'
        >>> filepath3 = 'path/of/file'
        >>> join_path(filepath1, filepath2, filepath3)
        '/path/of/dir/dir2/path/of/file'
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    return backend.join_path(filepath, *filepaths)


@contextmanager
def get_local_path(
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Generator[Union[str, Path], None, None]:
    """Download data from ``filepath`` and write the data to local path.

    ``get_local_path`` is decorated by :meth:`contxtlib.contextmanager`. It
    can be called with ``with`` statement, and when exists from the
    ``with`` statement, the temporary path will be released.

    Note:
        If the ``filepath`` is a local path, just return itself and it will
        not be released (removed).

    Args:
        filepath (str or Path): Path to be read data.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Yields:
        Iterable[str]: Only yield one path.

    Examples:
        >>> with get_local_path('s3://bucket/abc.jpg') as path:
        ...     # do something here
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    with backend.get_local_path(str(filepath)) as local_path:
        yield local_path


def copyfile(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    """Copy a file src to dst and return the destination file.

    src and dst should have the same prefix. If dst specifies a directory,
    the file will be copied into dst using the base filename from src. If
    dst specifies a file that already exists, it will be replaced.

    Args:
        src (str or Path): A file to be copied.
        dst (str or Path): Copy file to dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        str: The destination file.

    Raises:
        SameFileError: If src and dst are the same file, a SameFileError will
            be raised.

    Examples:
        >>> # dst is a file
        >>> src = '/path/of/file'
        >>> dst = '/path1/of/file1'
        >>> # src will be copied to '/path1/of/file1'
        >>> copyfile(src, dst)
        '/path1/of/file1'

        >>> # dst is a directory
        >>> dst = '/path1/of/dir'
        >>> # src will be copied to '/path1/of/dir/file'
        >>> copyfile(src, dst)
        '/path1/of/dir/file'
    """
    backend = get_file_backend(
        src, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copyfile(src, dst)


def copytree(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    """Recursively copy an entire directory tree rooted at src to a directory
    named dst and return the destination directory.

    src and dst should have the same prefix and dst must not already exist.

    Args:
        src (str or Path): A directory to be copied.
        dst (str or Path): Copy directory to dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.
        backend_key (str, optional): The key to get the backend from register.

    Returns:
        str: The destination directory.

    Raises:
        FileExistsError: If dst had already existed, a FileExistsError will be
            raised.

    Examples:
        >>> src = '/path/of/dir1'
        >>> dst = '/path/of/dir2'
        >>> copytree(src, dst)
        '/path/of/dir2'
    """
    backend = get_file_backend(
        src, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copytree(src, dst)


def copyfile_from_local(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    """Copy a local file src to dst and return the destination file.

    Note:
        If the backend is the instance of LocalBackend, it does the same
        thing with :func:`copyfile`.

    Args:
        src (str or Path): A local file to be copied.
        dst (str or Path): Copy file to dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        str: If dst specifies a directory, the file will be copied into dst
        using the base filename from src.

    Examples:
        >>> # dst is a file
        >>> src = '/path/of/file'
        >>> dst = 's3://openmmlab/mmengine/file1'
        >>> # src will be copied to 's3://openmmlab/mmengine/file1'
        >>> copyfile_from_local(src, dst)
        s3://openmmlab/mmengine/file1

        >>> # dst is a directory
        >>> dst = 's3://openmmlab/mmengine'
        >>> # src will be copied to 's3://openmmlab/mmengine/file''
        >>> copyfile_from_local(src, dst)
        's3://openmmlab/mmengine/file'
    """
    backend = get_file_backend(
        dst, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copyfile_from_local(src, dst)


def copytree_from_local(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    """Recursively copy an entire directory tree rooted at src to a directory
    named dst and return the destination directory.

    Note:
        If the backend is the instance of LocalBackend, it does the same
        thing with :func:`copytree`.

    Args:
        src (str or Path): A local directory to be copied.
        dst (str or Path): Copy directory to dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        str: The destination directory.

    Examples:
        >>> src = '/path/of/dir'
        >>> dst = 's3://openmmlab/mmengine/dir'
        >>> copyfile_from_local(src, dst)
        's3://openmmlab/mmengine/dir'
    """
    backend = get_file_backend(
        dst, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copytree_from_local(src, dst)


def copyfile_to_local(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    """Copy the file src to local dst and return the destination file.

    If dst specifies a directory, the file will be copied into dst using
    the base filename from src. If dst specifies a file that already
    exists, it will be replaced.

    Note:
        If the backend is the instance of LocalBackend, it does the same
        thing with :func:`copyfile`.

    Args:
        src (str or Path): A file to be copied.
        dst (str or Path): Copy file to to local dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        str: If dst specifies a directory, the file will be copied into dst
        using the base filename from src.

    Examples:
        >>> # dst is a file
        >>> src = 's3://openmmlab/mmengine/file'
        >>> dst = '/path/of/file'
        >>> # src will be copied to '/path/of/file'
        >>> copyfile_to_local(src, dst)
        '/path/of/file'

        >>> # dst is a directory
        >>> dst = '/path/of/dir'
        >>> # src will be copied to '/path/of/dir/file'
        >>> copyfile_to_local(src, dst)
        '/path/of/dir/file'
    """
    Path(dst).parent.mkdir(parents=True, exist_ok=True)
    backend = get_file_backend(
        src, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copyfile_to_local(src, dst)


def copytree_to_local(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Union[str, Path]:
    """Recursively copy an entire directory tree rooted at src to a local
    directory named dst and return the destination directory.

    Note:
        If the backend is the instance of LocalBackend, it does the same
        thing with :func:`copytree`.

    Args:
        src (str or Path): A directory to be copied.
        dst (str or Path): Copy directory to local dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        str: The destination directory.

    Examples:
        >>> src = 's3://openmmlab/mmengine/dir'
        >>> dst = '/path/of/dir'
        >>> copytree_to_local(src, dst)
        '/path/of/dir'
    """
    Path(dst).parent.mkdir(parents=True, exist_ok=True)
    backend = get_file_backend(
        dst, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copytree_to_local(src, dst)


def remove(
    filepath: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> None:
    """Remove a file.

    Args:
        filepath (str, Path): Path to be removed.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Raises:
        FileNotFoundError: If filepath does not exist, an FileNotFoundError
            will be raised.
        IsADirectoryError: If filepath is a directory, an IsADirectoryError
            will be raised.

    Examples:
        >>> filepath = '/path/of/file'
        >>> remove(filepath)
    """
    backend = get_file_backend(
        filepath,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    backend.remove(filepath)


def rmtree(
    dir_path: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> None:
    """Recursively delete a directory tree.

    Args:
        dir_path (str or Path): A directory to be removed.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Examples:
        >>> dir_path = '/path/of/dir'
        >>> rmtree(dir_path)
    """
    backend = get_file_backend(
        dir_path,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    backend.rmtree(dir_path)


def copy_if_symlink_fails(
    src: Union[str, Path],
    dst: Union[str, Path],
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> bool:
    """Create a symbolic link pointing to src named dst.

    If failed to create a symbolic link pointing to src, directory copy src to
    dst instead.

    Args:
        src (str or Path): Create a symbolic link pointing to src.
        dst (str or Path): Create a symbolic link named dst.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        bool: Return True if successfully create a symbolic link pointing to
        src. Otherwise, return False.

    Examples:
        >>> src = '/path/of/file'
        >>> dst = '/path1/of/file1'
        >>> copy_if_symlink_fails(src, dst)
        True
        >>> src = '/path/of/dir'
        >>> dst = '/path1/of/dir1'
        >>> copy_if_symlink_fails(src, dst)
        True
    """
    backend = get_file_backend(
        src, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.copy_if_symlink_fails(src, dst)


def list_dir_or_file(
    dir_path: Union[str, Path],
    list_dir: bool = True,
    list_file: bool = True,
    suffix: Optional[Union[str, Tuple[str]]] = None,
    recursive: bool = False,
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> Iterator[str]:
    """Scan a directory to find the interested directories or files in
    arbitrary order.

    Note:
        :meth:`list_dir_or_file` returns the path relative to ``dir_path``.

    Args:
        dir_path (str or Path): Path of the directory.
        list_dir (bool): List the directories. Defaults to True.
        list_file (bool): List the path of files. Defaults to True.
        suffix (str or tuple[str], optional): File suffix that we are
            interested in. Defaults to None.
        recursive (bool): If set to True, recursively scan the directory.
            Defaults to False.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Yields:
        Iterable[str]: A relative path to ``dir_path``.

    Examples:
        >>> dir_path = '/path/of/dir'
        >>> for file_path in list_dir_or_file(dir_path):
        ...     print(file_path)
        >>> # list those files and directories in current directory
        >>> for file_path in list_dir_or_file(dir_path):
        ...     print(file_path)
        >>> # only list files
        >>> for file_path in list_dir_or_file(dir_path, list_dir=False):
        ...     print(file_path)
        >>> # only list directories
        >>> for file_path in list_dir_or_file(dir_path, list_file=False):
        ...     print(file_path)
        >>> # only list files ending with specified suffixes
        >>> for file_path in list_dir_or_file(dir_path, suffix='.txt'):
        ...     print(file_path)
        >>> # list all files and directory recursively
        >>> for file_path in list_dir_or_file(dir_path, recursive=True):
        ...     print(file_path)
    """
    backend = get_file_backend(
        dir_path,
        backend_args=backend_args,
        enable_singleton=True,
        backend_key=backend_key,
    )
    yield from backend.list_dir_or_file(
        dir_path, list_dir, list_file, suffix, recursive
    )


def generate_presigned_url(
    url: str,
    client_method: str = "get_object",
    expires_in: int = 3600,
    backend_args: Optional[dict] = None,
    backend_key: Optional[str] = None,
) -> str:
    """Generate the presigned url of video stream which can be passed to
    mmcv.VideoReader. Now only work on s3 backend.

    Note:
        Now only work on s3 backend.

    Args:
        url (str): Url of video stream.
        client_method (str): Method of client, 'get_object' or
            'put_object'. Defaults to 'get_object'.
        expires_in (int): expires, in seconds. Defaults to 3600.
        backend_args (dict, optional): Arguments to instantiate the
            corresponding backend. Defaults to None.

    Returns:
        str: Generated presigned url.
    """
    backend = get_file_backend(
        url, backend_args=backend_args, enable_singleton=True, backend_key=backend_key
    )
    return backend.generate_presigned_url(url, client_method, expires_in)


def load(
    file,
    file_format=None,
    file_client_args=None,
    backend_args=None,
    backend_key=None,
    **kwargs,
):
    """Load data from json/yaml/pickle files.

    This method provides a unified api for loading data from serialized files.

    ``load`` supports loading data from serialized files those can be storaged
    in different backends.

    Args:
        file (str or :obj:`Path` or file-like object): Filename or a file-like
            object.
        file_format (str, optional): If not specified, the file format will be
            inferred from the file extension, otherwise use the specified one.
            Currently supported formats include "json", "yaml/yml" and
            "pickle/pkl".
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmengine.fileio.FileClient` for details.
            Defaults to None. It will be deprecated in future. Please use
            ``backend_args`` instead.
        backend_args (dict, optional): Arguments to instantiate the
            prefix of uri corresponding backend. Defaults to None.
            New in v0.2.0.

    Examples:
        >>> load('/path/of/your/file')  # file is storaged in disk
        >>> load('https://path/of/your/file')  # file is storaged in Internet
        >>> load('s3://path/of/your/file')  # file is storaged in s3

    Returns:
        The content from the file.
    """
    if isinstance(file, Path):
        file = str(file)
    if file_format is None and isinstance(file, str):
        file_format = file.split(".")[-1]
    if file_format not in file_handlers:
        raise TypeError(f"Unsupported format: {file_format}")

    if file_client_args is not None:
        warnings.warn(
            '"file_client_args" will be deprecated in future. '
            'Please use "backend_args" instead',
            DeprecationWarning,
        )
        if backend_args is not None:
            raise ValueError(
                '"file_client_args and "backend_args" cannot be set at the '
                "same time."
            )

    handler = file_handlers[file_format]
    if isinstance(file, str):
        if file_client_args is not None:
            file_client = FileClient.infer_client(file_client_args, file)
            file_backend = file_client
        else:
            file_backend = get_file_backend(
                file,
                backend_args=backend_args,
                backend_key=backend_key,
                enable_singleton=True,
            )

        if handler.str_like:
            with StringIO(file_backend.get_text(file)) as f:
                obj = handler.load_from_fileobj(f, **kwargs)
        else:
            with BytesIO(file_backend.get(file)) as f:
                obj = handler.load_from_fileobj(f, **kwargs)
    elif hasattr(file, "read"):
        obj = handler.load_from_fileobj(file, **kwargs)
    else:
        raise TypeError('"file" must be a filepath str or a file-object')
    return obj


def dump(
    obj,
    file=None,
    file_format=None,
    file_client_args=None,
    backend_args=None,
    backend_key=None,
    **kwargs,
):
    """Dump data to json/yaml/pickle strings or files.

    This method provides a unified api for dumping data as strings or to files,
    and also supports custom arguments for each file format.

    ``dump`` supports dumping data as strings or to files which is saved to
    different backends.

    Args:
        obj (any): The python object to be dumped.
        file (str or :obj:`Path` or file-like object, optional): If not
            specified, then the object is dumped to a str, otherwise to a file
            specified by the filename or file-like object.
        file_format (str, optional): Same as :func:`load`.
        file_client_args (dict, optional): Arguments to instantiate a
            FileClient. See :class:`mmengine.fileio.FileClient` for details.
            Defaults to None. It will be deprecated in future. Please use
            ``backend_args`` instead.
        backend_args (dict, optional): Arguments to instantiate the
            prefix of uri corresponding backend. Defaults to None.
            New in v0.2.0.

    Examples:
        >>> dump('hello world', '/path/of/your/file')  # disk
        >>> dump('hello world', 's3://path/of/your/file')  # ceph or s3

    Returns:
        bool: True for success, False otherwise.
    """
    if isinstance(file, Path):
        file = str(file)
    if file_format is None:
        if isinstance(file, str):
            file_format = file.split(".")[-1]
        elif file is None:
            raise ValueError("file_format must be specified since file is None")
    if file_format not in file_handlers:
        raise TypeError(f"Unsupported format: {file_format}")

    if file_client_args is not None:
        warnings.warn(
            '"file_client_args" will be deprecated in future. '
            'Please use "backend_args" instead',
            DeprecationWarning,
        )
        if backend_args is not None:
            raise ValueError(
                '"file_client_args" and "backend_args" cannot be set at the '
                "same time."
            )

    handler = file_handlers[file_format]
    if file is None:
        return handler.dump_to_str(obj, **kwargs)
    elif isinstance(file, str):
        if file_client_args is not None:
            file_client = FileClient.infer_client(file_client_args, file)
            file_backend = file_client
        else:
            file_backend = get_file_backend(
                file,
                backend_args=backend_args,
                backend_key=backend_key,
                enable_singleton=True,
            )

        if handler.str_like:
            with StringIO() as f:
                handler.dump_to_fileobj(obj, f, **kwargs)
                file_backend.put_text(f.getvalue(), file)
        else:
            with BytesIO() as f:
                handler.dump_to_fileobj(obj, f, **kwargs)
                file_backend.put(f, file)
    elif hasattr(file, "write"):
        handler.dump_to_fileobj(obj, file, **kwargs)
    else:
        raise TypeError('"file" must be a filename str or a file-object')



import socket

__all__ = ["find_free_port", "get_local_addr", "is_port_used"]


def find_free_port():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]
    finally:
        s.close()


# http://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
def get_local_addr():
    try:
        resolve = [
            ip
            for ip in socket.gethostbyname_ex(socket.gethostname())[2]
            if not ip.startswith("127.")
        ][:1]
        if len(resolve):
            return resolve[0]
        # AF_INET: Address family, ipv4. SOCK_DGRAM: connections, unreliable, datagrams UDP
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        addr = s.getsockname()[0]
        s.close()
        return addr
    except Exception:  # pylint: disable=broad-except
        try:
            return socket.gethostbyname(socket.gethostname())
        except Exception:  # pylint: disable=broad-except
            return "127.0.0.1"


def is_port_used(port: int):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : meta.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/18/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import collections
import contextlib
import functools
import operator
import threading

import six

__all__ = [
    "gofor",
    "run_once",
    "try_run",
    "map_exec",
    "filter_exec",
    "first_n",
    "stmap",
    "method2func",
    "map_exec_method",
    "decorator_with_optional_args",
    "cond_with",
    "cond_with_group",
    "merge_iterable",
    "dict_deep_update",
    "dict_deep_kv",
    "dict_deep_keys",
    "assert_instance",
    "assert_none",
    "assert_notnone",
    "notnone_property",
    "synchronized",
    "make_dummy_func",
    "Singleton",
]


def gofor(v):
    if isinstance(v, collections.Mapping):
        return v.items()
    assert_instance(v, collections.Iterable)
    return enumerate(v)


def run_once(func):
    has_run = False

    @synchronized
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        nonlocal has_run
        if not has_run:
            has_run = True
            return func(*args, **kwargs)
        else:
            return

    return new_func


def try_run(lambda_):
    try:
        return lambda_()
    except Exception:
        return None


def map_exec(func, *iterables):
    return list(map(func, *iterables))


def filter_exec(func, iterable):
    return list(filter(func, iterable))


def first_n(iterable, n=10):
    def gen():
        it = iter(iterable)
        for i in range(n):
            try:
                yield next(it)
            except StopIteration:
                return

    return list(gen())


def stmap(func, iterable):
    if isinstance(iterable, six.string_types):
        return func(iterable)
    if isinstance(iterable, (collections.abc.Sequence, collections.UserList)):
        return [stmap(func, v) for v in iterable]
    if isinstance(iterable, collections.abc.Set):
        return {stmap(func, v) for v in iterable}
    if isinstance(iterable, (collections.abc.Mapping, collections.UserDict)):
        return {k: stmap(func, v) for k, v in iterable.items()}

    return func(iterable)


def method2func(method_name):
    return lambda x: getattr(x, method_name)()


def map_exec_method(method_name, iterable):
    return list(map(method2func(method_name), iterable))


def decorator_with_optional_args(func=None, *, is_method=False):
    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            if is_method:
                if len(args) == 1:
                    return f(*args, **kwargs)
                elif len(args) == 2:
                    return f(args[0], **kwargs)(args[1])
                else:
                    raise ValueError(
                        "Decorator supports 0 or 1 positional arguments as the function to be wrapped."
                    )
            else:
                if len(args) == 0:
                    return f(**kwargs)
                elif len(args) == 1:
                    return f(**kwargs)(args[0])
                else:
                    raise ValueError(
                        "Decorator supports 0 or 1 positional arguments as the function to be wrapped."
                    )

        return wrapped

    if func is not None:
        return wrapper(func)
    else:
        return wrapper


@contextlib.contextmanager
def cond_with(with_statement, cond):
    if cond:
        with with_statement as res:
            yield res
    else:
        yield


@contextlib.contextmanager
def cond_with_group(cond, *with_statement):
    if cond:
        with contextlib.ExitStack() as stack:
            res = [stack.enter_context(ctx) for ctx in with_statement]
            yield res
    else:
        yield


def merge_iterable(v1, v2):
    assert issubclass(type(v1), type(v2)) or issubclass(type(v2), type(v1))
    if isinstance(v1, (dict, set)):
        v = v1.copy().update(v2)
        return v

    return v1 + v2


def dict_deep_update(a, b):
    for key in b:
        if key in a and type(b[key]) is dict:
            dict_deep_update(a[key], b[key])
        else:
            a[key] = b[key]


def dict_deep_kv(d, sort=True, sep=".", allow_dict=False):
    # Not using collections.Sequence to avoid infinite recursion.
    assert isinstance(d, (tuple, list, collections.Mapping))
    result = list()

    def _dfs(current, prefix=None):
        for key, value in gofor(current):
            current_key = key if prefix is None else prefix + sep + str(key)
            if isinstance(current[key], (tuple, list, collections.Mapping)):
                if allow_dict:
                    result.append((current_key, value))
                _dfs(current[key], current_key)
            else:
                result.append((current_key, value))

    _dfs(d)
    if sort:
        result.sort(key=operator.itemgetter(0))
    return result


def dict_deep_keys(d, sort=True, sep=".", allow_dict=True):
    kv = dict_deep_kv(d, sort=sort, sep=sep, allow_dict=allow_dict)
    return [i[0] for i in kv]


def assert_instance(ins, clz, msg=None):
    msg = msg or "{} (of type{}) is not of type {}".format(ins, type(ins), clz)
    assert isinstance(ins, clz), msg


def assert_none(ins, msg=None):
    msg = msg or "{} is not None".format(ins)
    assert ins is None, msg


def assert_notnone(ins, msg=None, name="instance"):
    msg = msg or "{} is None".format(name)
    assert ins is not None, msg


class notnone_property:
    def __init__(self, fget):
        self.fget = fget
        self.__module__ = fget.__module__
        self.__name__ = fget.__name__
        self.__doc__ = fget.__doc__
        self.__prop_key = "{}_{}".format(fget.__name__, id(fget))

    def __get__(self, instance, owner):
        if instance is None:
            return self.fget
        v = self.fget(instance)
        assert v is not None, "{}.{} can not be None, maybe not set yet".format(
            type(instance).__name__, self.__name__
        )
        return v


@decorator_with_optional_args
def synchronized(mutex=None):
    if mutex is None:
        mutex = threading.Lock()

    def wrapper(func):
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            with mutex:
                return func(*args, **kwargs)

        wrapped_func.__sync_mutex__ = mutex
        return wrapped_func

    return wrapper


def make_dummy_func(message=None):
    def func(*args, **kwargs):
        raise NotImplementedError(message)

    return func


class Singleton(type):
    "https://stackoverflow.com/a/55629949/7672954"
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._locked_call(*args, **kwargs)
        return cls._instances[cls]

    @synchronized
    def _locked_call(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

    @property
    def ready(cls):
        if cls in cls._instances:
            return True
        return False

    @classmethod
    def clear(cls):
        cls._instances.clear()


def repr_from_str(self):
    """A helper function to generate the repr string from the __str__ method.

    Example:
        .. code-block:: python

            class Foo(object):
                def __str__(self):
                    return 'Foo'

                __repr__ = repr_from_str

            print(Foo())  # Foo
            print(repr(Foo()))  # Foo[Foo]
    """
    return "{}<{}>".format(self.__class__.__name__, self.__str__())



import tarfile

from .base import BaseFileHandler


class TarHandler(BaseFileHandler):
    str_like = False

    def load_from_fileobj(self, file, mode="r|*", **kwargs):
        return tarfile.open(fileobj=file, mode=mode, **kwargs)

    def load_from_path(self, filepath, mode="r|*", **kwargs):
        return tarfile.open(filepath, mode=mode, **kwargs)

    def dump_to_fileobj(self, obj, file, mode="w", **kwargs):
        with tarfile.open(fileobj=file, mode=mode) as tar:
            tar.add(obj, **kwargs)

    def dump_to_path(self, obj, filepath, mode="w", **kwargs):
        with tarfile.open(filepath, mode=mode) as tar:
            tar.add(obj, **kwargs)

    def dump_to_str(self, obj, **kwargs):
        raise NotImplementedError



import functools
from types import FunctionType

class LazyDict(dict):
    def __getitem__(self, key):
        obj = dict.__getitem__(self, key)
        if isinstance(obj, FunctionType) or isinstance(obj, functools.partial):
            obj = obj()
            dict.__setitem__(self, key, obj)
        return obj


#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : registry.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 12/29/16
#
# Distributed under terms of the MIT license.

import collections
import threading

from .context import EmptyContext

__all__ = [
    "Registry",
    "DefaultRegistry",
    "RegistryGroup",
    "CallbackRegistry",
    "LockRegistry",
]


class Registry:
    __FALLBACK_KEY__ = "__fallback__"

    _registry = None

    def __init__(self):
        self._init_registry()

    def _init_registry(self):
        self._registry = {}

    @property
    def fallback(self):
        return self._registry.get(self.__FALLBACK_KEY__, None)

    def set_fallback(self, value):
        self._registry[self.__FALLBACK_KEY__] = value
        return self

    def register(self, entry, value):
        self._registry[entry] = value
        return self

    def unregister(self, entry):
        return self._registry.pop(entry, None)

    def has(self, entry):
        return entry in self._registry

    def lookup(self, entry, fallback=True, default=None):
        if fallback:
            fallback_value = self._registry.get(self.__FALLBACK_KEY__, default)
        else:
            fallback_value = default
        return self._registry.get(entry, fallback_value)

    def keys(self):
        return list(self._registry.keys())

    def items(self):
        return list(self._registry.items())


class DefaultRegistry(Registry):
    __base_class__ = dict

    def _init_registry(self):
        base_class = type(self).__base_class__
        self._registry = collections.defaultdict(base_class)

    def lookup(self, entry, fallback=False, default=None):
        assert fallback is False and default is None
        return self._registry[entry]

    def __getitem__(self, item):
        return self.lookup(item)


class RegistryGroup:
    __base_class__ = Registry

    def __init__(self):
        self._init_registry_group()

    def _init_registry_group(self):
        base_class = type(self).__base_class__
        self._registries = collections.defaultdict(base_class)

    def __getitem__(self, item):
        return self._registries[item]

    def register(self, registry_name, entry, value, **kwargs):
        return self._registries[registry_name].register(entry, value, **kwargs)

    def lookup(self, registry_name, entry, fallback=True, default=None):
        return self._registries[registry_name].lookup(
            entry, fallback=fallback, default=default
        )


class CallbackRegistry(Registry):
    """
    A callable manager utils.

    If there exists a super callback, it will block all callbacks.
    A super callback will receive the called name as its first argument.

    Then the dispatcher will try to call the callback by name.
    If such name does not exists, a fallback callback will be called.

    The fallback callback will also receive the called name as its first argument.

    Examples:

    >>> registry = CallbackRegistry()
    >>> callback_func = print
    >>> registry.register('name', callback_func)  # register a callback.
    >>> registry.dispatch('name', 'arg1', 'arg2', kwarg1='kwarg1')  # dispatch.
    """

    def __init__(self):
        super().__init__()
        self._super_callback = None

    @property
    def super_callback(self):
        return self._super_callback

    def set_super_callback(self, callback):
        self._super_callback = callback
        return self

    @property
    def fallback_callback(self):
        return self.fallback

    def set_fallback_callback(self, callback):
        return self.set_fallback(callback)

    def dispatch(self, name, *args, **kwargs):
        if self._super_callback is not None:
            return self._super_callback(self, name, *args, **kwargs)
        return self.dispatch_direct(name, *args)

    def dispatch_direct(self, name, *args, **kwargs):
        """Dispatch by name, ignoring the super callback."""
        callback = self.lookup(name, fallback=False)
        if callback is None:
            if self.fallback_callback is None:
                raise ValueError('Unknown callback entry: "{}".'.format(name))
            return self.fallback_callback(name, *args, **kwargs)
        return callback(*args, **kwargs)


class LockRegistry(DefaultRegistry):
    __base_class__ = threading.Lock

    def synchronized(self, entry, activate=True):
        if activate:
            return self.lookup(entry)
        return EmptyContext()



"""Implementations of Normal distributions."""

import numpy as np
import torch
from torch import nn

from jamtorch.utils import ts_ops

from .base import Distribution


class StandardNormal(Distribution):
    """A multivariate Normal with zero mean and unit covariance."""

    def __init__(self, shape):
        super().__init__()
        self._shape = torch.Size(shape)

        self.register_buffer(
            "_log_z",
            torch.tensor(0.5 * np.prod(shape) * np.log(2 * np.pi), dtype=torch.float64),
            persistent=False,
        )

    def _log_prob(self, inputs, context):
        # Note: the context is ignored.
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )
        neg_energy = -0.5 * ts_ops.sum_except_batch(inputs ** 2, num_batch_dims=1)
        return neg_energy - self._log_z

    def _sample(self, num_samples, context):
        if context is None:
            return torch.randn(num_samples, *self._shape, device=self._log_z.device)
        else:
            # The value of the context is ignored, only its size and device are taken into account.
            context_size = context.shape[0]
            samples = torch.randn(
                context_size * num_samples, *self._shape, device=context.device
            )
            return ts_ops.split_leading_dim(samples, [context_size, num_samples])

    def _mean(self, context):
        if context is None:
            return self._log_z.new_zeros(self._shape)
        else:
            # The value of the context is ignored, only its size is taken into account.
            return context.new_zeros(context.shape[0], *self._shape)


class ConditionalDiagonalNormal(Distribution):
    """A diagonal multivariate Normal whose parameters are functions of a context."""

    def __init__(self, shape, context_encoder=None):
        """Constructor.

        Args:
            shape: list, tuple or torch.Size, the shape of the input variables.
            context_encoder: callable or None, encodes the context to the distribution parameters.
                If None, defaults to the identity function.
        """
        super().__init__()
        self._shape = torch.Size(shape)
        if context_encoder is None:
            self._context_encoder = lambda x: x
        else:
            self._context_encoder = context_encoder
        self.register_buffer(
            "_log_z",
            torch.tensor(0.5 * np.prod(shape) * np.log(2 * np.pi), dtype=torch.float64),
            persistent=False,
        )

    def _compute_params(self, context):
        """Compute the means and log stds form the context."""
        if context is None:
            raise ValueError("Context can't be None.")

        params = self._context_encoder(context)
        if params.shape[-1] % 2 != 0:
            raise RuntimeError(
                "The context encoder must return a tensor whose last dimension is even."
            )
        if params.shape[0] != context.shape[0]:
            raise RuntimeError(
                "The batch dimension of the parameters is inconsistent with the input."
            )

        split = params.shape[-1] // 2
        means = params[..., :split].reshape(params.shape[0], *self._shape)
        log_stds = params[..., split:].reshape(params.shape[0], *self._shape)
        return means, log_stds

    def _log_prob(self, inputs, context):
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )

        # Compute parameters.
        means, log_stds = self._compute_params(context)
        assert means.shape == inputs.shape and log_stds.shape == inputs.shape

        # Compute log prob.
        norm_inputs = (inputs - means) * torch.exp(-log_stds)
        log_prob = -0.5 * ts_ops.sum_except_batch(norm_inputs ** 2, num_batch_dims=1)
        log_prob -= ts_ops.sum_except_batch(log_stds, num_batch_dims=1)
        log_prob -= self._log_z
        return log_prob

    def _sample(self, num_samples, context):
        # Compute parameters.
        means, log_stds = self._compute_params(context)
        stds = torch.exp(log_stds)
        means = ts_ops.repeat_rows(means, num_samples)
        stds = ts_ops.repeat_rows(stds, num_samples)

        # Generate samples.
        context_size = context.shape[0]
        noise = torch.randn(
            context_size * num_samples, *self._shape, device=means.device
        )
        samples = means + stds * noise
        return ts_ops.split_leading_dim(samples, [context_size, num_samples])

    def _mean(self, context):
        means, _ = self._compute_params(context)
        return means


class DiagonalNormal(Distribution):
    """A diagonal multivariate Normal with trainable parameters."""

    def __init__(self, shape, mean=None, std=None):
        """Constructor.

        Args:
            shape: list, tuple or torch.Size, the shape of the input variables.
            context_encoder: callable or None, encodes the context to the distribution parameters.
                If None, defaults to the identity function.
        """
        super().__init__()
        self._shape = torch.Size(shape)
        if mean is None:
            self.mean_ = nn.Parameter(torch.zeros(shape).reshape(1, -1))
        else:
            assert isinstance(mean, torch.Tensor)
            assert np.prod(shape) == mean.nelement()
            self.mean_ = nn.Parameter(mean.view(1, -1))
        if std is None:
            self.log_std_ = nn.Parameter(torch.zeros(shape).reshape(1, -1))
        else:
            assert isinstance(std, torch.Tensor)
            assert np.prod(shape) == std.nelement()
            self.log_std_ = nn.Parameter(torch.log(std.view(1, -1)))

    def _log_prob(self, inputs, context):
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )

        # Compute parameters.
        means = self.mean_
        log_stds = self.log_std_

        # Compute log prob.
        norm_inputs = (inputs - means) * torch.exp(-log_stds)
        log_prob = -0.5 * ts_ops.sum_except_batch(norm_inputs ** 2, num_batch_dims=1)
        # log_prob -= ts_ops.sum_except_batch(log_stds, num_batch_dims=1)
        log_prob -= self._log_z()
        return log_prob

    def _sample(self, num_samples, context):
        raise NotImplementedError()

    def _mean(self, context):
        return self.mean

    def _log_z(self):
        const = 0.5 * np.prod(self._shape) * np.log(2 * np.pi)
        return const + torch.log(torch.prod(torch.exp(self.log_std_)))



import os
import stackprinter
from filelock import FileLock, SoftFileLock

# pylint: disable=broad-except

__all__ = [
    "try_filelock",
    "TryFileLock"
]


def try_filelock(
    fn,
    target_file,
    is_soft=False,
    timeout=20,
):
    lock_obj = SoftFileLock if is_soft else FileLock
    try:
        os.makedirs(os.path.dirname(target_file), exist_ok=True)
        with lock_obj(f"{target_file}._lock", timeout) as flock:
            if flock.is_locked:
                fn()
            else:
                print(f"try_filelock Timeout!!! {target_file}")
    except TimeoutError:
        print(f"TryFileLock timeout!!! {target_file}")
    except Exception as exc:
        stackprinter.show(exc, style="darkbg2")

class TryFileLock:
    def __init__(self, target_file, is_soft=False, timeout=20):
        self.target_file = target_file
        self.is_soft = is_soft
        self.timeout = timeout
        self.lock_obj = SoftFileLock if is_soft else FileLock
        self.lock = None

    def __enter__(self):
        try:
            os.makedirs(os.path.dirname(self.target_file), exist_ok=True)
            self.lock = self.lock_obj(f"{self.target_file}._lock", self.timeout)
            self.lock.acquire()
            flag = self.lock.is_locked
            if flag:
                return self
            else:
                print(f"TryFileLock can not acquire lock {self.target_file}")
                return None
        # handle timeout error
        except TimeoutError:
            # raise TimeoutError(f"TryFileLock timeout!!! {self.target_file}")
            print(f"TryFileLock timeout!!! {self.target_file}")
            return None
        except Exception as exc:
            stackprinter.show(exc, style="darkbg2")
            raise RuntimeError(f"TryFileLock error!!!")

    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_tb
        if exc_type is not None:
            print(f"TryFileLock error!!! {exc_type} {exc_val}")
        if self.lock is not None and self.lock.is_locked:
            self.lock.release()


from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
import os.path as osp
import torch.optim as optim
from torchvision import datasets, transforms
from torch.optim.lr_scheduler import StepLR
from jamtorch.trainer import step_lr
from jamtorch.trainer.simple_trainer import SimpleTrainer
from jammy.logging import Wandb
import hydra
from omegaconf import OmegaConf


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


# TODO: should consider process after eval
def build_loss(device):
    def loss_fn(model, feed_dict, is_train, device=device):
        data, target = feed_dict
        data, target = data.to(device), target.to(device)
        output = model(data)
        loss = F.nll_loss(output, target)
        if is_train:
            return loss, {}, {}

        pred = output.argmax(dim=1, keepdim=True)
        correct = 1.0 * pred.eq(target.view_as(pred)).sum() / len(pred)
        return loss, {}, {"correct": correct}

    return loss_fn


def run(cfg):
    use_cuda = not cfg.no_cuda and torch.cuda.is_available()
    if use_cuda:
        print("using CUDA")

    device = torch.device("cuda" if use_cuda else "cpu")

    train_kwcfg = {"batch_size": cfg.data.batch_size}
    test_kwcfg = {"batch_size": cfg.data.test_batch_size}
    if use_cuda:
        cuda_kwargs = {"num_workers": 1, "pin_memory": True, "shuffle": True}
        train_kwcfg.update(cuda_kwargs)
        test_kwcfg.update(cuda_kwargs)

    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    data_path = osp.join(hydra.utils.get_original_cwd(), "data")
    dataset1 = datasets.MNIST(data_path, train=True, download=True, transform=transform)
    dataset2 = datasets.MNIST(data_path, train=False, transform=transform)
    train_loader = torch.utils.data.DataLoader(dataset1, **train_kwcfg)
    test_loader = torch.utils.data.DataLoader(dataset2, **test_kwcfg)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=cfg.lr)

    scheduler = StepLR(optimizer, step_size=1, gamma=cfg.gamma)

    trainer = SimpleTrainer(cfg.trainer, build_loss(device))
    trainer.set_model_optim(model, optimizer)
    trainer.lr_scheduler = scheduler
    trainer.set_monitor(cfg.trainer.wandb.log)
    trainer.set_dataloader(train_loader, test_loader)
    trainer.register_event("epoch:after", step_lr)
    trainer.train()
    # trainer(train_loader, test_loader)


@hydra.main(config_name="mnist.yaml")
def main(cfg):
    Wandb.launch(cfg, cfg.trainer.wandb.log, True)
    OmegaConf.set_struct(cfg, False)
    run(cfg)
    Wandb.finish()


if __name__ == "__main__":
    main()



import torch
import numpy
import os
from jammy.utils.registry import Registry
# from jammy.logging import get_logger

def worker():
    pid = os.getpid()
    registry = Registry()
    registry.register("torch", lambda: torch.rand(4))
    registry.register("numpy", lambda: numpy.random.randn(4))
    for k, fn in registry.items():
        print(f"{pid:05d}\t {k}\t {fn()}")



import torch
from jamtorch.logging import get_logger
import torch.nn as nn

__all__ = ["ParamEMA"]

logger = get_logger()


class ParamEMA:
    def __init__(self, beta, num_warm, num_every, model=None, forget_resume=False):
        if beta < 0.0 or beta > 1.0:
            raise ValueError("beta must be between 0 and 1")
        self.one_minus_decay = 1 - beta
        self.cnt = 0
        self.num_warm = num_warm
        self.num_every = num_every
        self.shadow_params = None
        self.forget_resume = forget_resume

    def update_parameters(self, new_model):
        if hasattr(new_model, "module"):
            new_model = new_model.module
        self.cnt += 1
        if self.cnt < self.num_warm:
            return
        if self.shadow_params is None:
            logger.debug("Ema finish the initialization")
            self.shadow_params = [
                p.clone().detach().cpu() for p in new_model.parameters()
            ]
        with torch.no_grad():
            if self.cnt % self.num_every == 0:
                for s_param, param in zip(self.shadow_params, new_model.parameters()):
                    s_param.sub_(self.one_minus_decay * (s_param - param.cpu()))

    def copy_to(self, model):
        if hasattr(model, "module"):
            model = model.module
        for s_param, param in zip(self.shadow_params, model.parameters()):
            param.data.copy_(s_param.to(param.device).data)
        return model

    def state_dict(self):
        return {
            "one_minus_decay": self.one_minus_decay,
            "model": self.shadow_params,
            "num_warm": self.num_warm,
            "num_every": self.num_every,
            "cnt": self.cnt,
        }

    def load_state_dict(self, ema_dict):
        self.cnt = ema_dict.get("cnt") or 0
        self.num_warm = ema_dict.get("num_warm") or 0
        self.num_every = ema_dict.get("num_every") or 1
        if not self.forget_resume:
            self.shadow_params = ema_dict.get("model") or []
            if self.shadow_params:
                self.shadow_params = [param.cpu() for param in ema_dict.get("model")]
            # for param in self.shadow_params:
            # param.cpu()
        self.one_minus_decay = ema_dict.get("one_minus_decay") or 0.1
        logger.critical(
            f"=====> Loading EMA finish =====> with forget_resume: {self.forget_resume}"
        )



import datetime
import functools
import os
from contextlib import contextmanager, nullcontext

import torch
import torch.distributed as dist

import jammy.comm as comm

__all__ = [
    "is_master",
    "is_dist",
    "master_first",
    "master_only",
    "ddp_runner",
    "ddp_loaders",
    "get_world_size",
    "barrier",
]


def is_master():
    return not dist.is_initialized() or dist.get_rank() == 0


def is_dist():
    return dist.is_initialized()


def get_world_size():
    if is_dist():
        return dist.get_world_size()
    return 1


def barrier():
    if is_dist():
        dist.barrier()


@contextmanager
def master_first():
    if not is_master():
        barrier()
    yield
    if dist.is_initialized() and is_master():
        barrier()


def master_only(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):  # pylint: disable=inconsistent-return-statements
        if is_master():
            return func(*args, **kwargs)
        return

    return new_func


def ddp_setup(rank, world_size, working_dir, cfg):
    # TODO: discuss the necessarity?
    cfg.rank = rank
    cfg.gpu = rank
    cfg.cwd = working_dir
    cfg.world_size = world_size

    # different random seed for different process
    torch.manual_seed(rank)

    os.environ["MASTER_ADDR"] = cfg.dist.master_addr
    os.environ["MASTER_PORT"] = cfg.dist.master_port
    timeout_sec = 1800
    if cfg.dist.timeout is not None:
        os.environ["NCCL_BLOCKING_WAIT"] = "1"
        timeout_sec = cfg.dist.timeout
    timeout = datetime.timedelta(seconds=timeout_sec)

    torch.cuda.set_device(rank)
    dist.init_process_group(
        backend=cfg.dist.mode, rank=rank, world_size=world_size, timeout=timeout
    )
    dist.barrier()


def cleanup():
    dist.destroy_process_group()


def ddp_runner(func):
    @functools.wraps(func)
    def new_fn(rank, n_gpu, working_dir, config):
        ddp_setup(rank, n_gpu, working_dir, config.trainer)
        if config.trainer.dist.adjust_lr:
            if "optimizer" in config:
                config.optimizer.lr = config.optimizer.lr * config.trainer.world_size
        value = func(config)
        cleanup()
        return value

    return new_fn


def ddp_loaders(train_set, val_set, rank=None, world_size=None, **dl_kwargs):
    """instantiate dataloaders, deal with ddp and non-ddp, return None sampler

    :type train_set: torch.utils.data.Dataset
    :type val_set: torch.utils.data.Dataset
    :param rank: [description], defaults to None
    :type rank: [type], optional
    :param world_size: [description], defaults to None
    :type world_size: [type], optional
    :return: train_loader, train_sampler, val_loader, val_sampler
    """
    if not dist.is_initialized():
        # no dist trainining
        if train_set is not None:
            train_loader = torch.utils.data.DataLoader(
                train_set, shuffle=True, **dl_kwargs
            )
        else:
            train_loader = None
        if val_set is not None:
            val_loader = torch.utils.data.DataLoader(val_set, shuffle=True, **dl_kwargs)
        else:
            val_loader = None

        return train_loader, None, val_loader, None

    if train_set is not None:
        train_sampler = torch.utils.data.DistributedSampler(train_set, world_size, rank)
        train_loader = torch.utils.data.DataLoader(
            train_set, sampler=train_sampler, **dl_kwargs
        )

    else:
        train_sampler = None
        train_loader = False

    if val_set is not None:
        val_sampler = torch.utils.data.DistributedSampler(val_set, world_size, rank)
        val_loader = torch.utils.data.DataLoader(
            val_set, sampler=val_sampler, **dl_kwargs
        )
    else:
        val_sampler = None
        val_loader = None

    return train_loader, train_sampler, val_loader, val_sampler


def fast_acc_grad_loss(trainer, loss):
    is_step = trainer.iter_cnt % trainer.ratio_forback != 0
    sync_context = (
        trainer.model.no_sync if dist.is_initialized() and is_step else nullcontext
    )
    # avoid sync gradient
    with sync_context:
        loss.backward()
    if is_step:
        return True

    return False


def prepare_cfg(cfg):
    # find good ports
    free_port = comm.find_free_port()
    cfg.trainer.dist.master_port = str(free_port)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : sc_server.py
# Author : Jiayuan Mao, Qsh.Zh
# Email  : qsh.zh27@gmail.com
# Date   : 11/17/2021
#
# Qinsheng modifies based on Jacinle.
# Distributed under terms of the MIT license.

import time

from jammy.comm.cs import ServerPipe


def answer(pipe, identifier, inp):
    out = inp["a"] + inp["b"]
    pipe.send(identifier, dict(out=out))


def main():
    server = ServerPipe("server", mode="ipc")
    server.dispatcher.register("calc", answer)
    with server.activate():
        print("Client command:")
        print(
            "jam-run sc_client.py", *server.conn_info  # pylint: disable=not-an-iterable
        )
        while True:
            time.sleep(1)


if __name__ == "__main__":
    main()



from abc import abstractmethod

import pytorch_lightning as pl  # pylint: disable=unused-import
from pytorch_lightning import Callback


class EveryN(Callback):
    def __init__(self, every_n):
        self.every_n = every_n

    def on_batch_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if pl_module.global_step % self.every_n == 0:
            self.every_n_impl(trainer, pl_module)

    @abstractmethod
    def every_n_impl(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        ...



import numpy as np
from einops import rearrange
from torch.utils.data import Dataset

__all__ = [
    "Points2Dataset",
]

# pylint: disable=import-outside-toplevel, import-error, unused-argument, too-many-instance-attributes, too-many-arguments


def swissroll_generate_sample(N, noise=0.25):
    import sklearn.datasets as skd

    data = skd.make_swiss_roll(n_samples=N, noise=noise)[0]
    data = data.astype("float32")[:, [0, 2]]
    return data


def moon_generate_sample(N, noise=0.25):
    import sklearn.datasets as skd

    data = skd.make_moons(n_samples=N, noise=noise)[0]
    data = data.astype("float32")
    return data


def checkerboard_generate_sample(N, noise=0.25):
    x1 = np.random.rand(N) * 4 - 2
    x2_ = np.random.rand(N) - np.random.randint(0, 2, N) * 2
    x2 = x2_ + (np.floor(x1) % 2)
    return np.concatenate([x1[:, None], x2[:, None]], 1) * 2


def line_generate_sample(N, noise=0.25):
    assert noise <= 1.0
    cov = np.array([[1.0, 1 - noise], [1 - noise, 1.0]])
    mean = np.array([0.0, 0.0])
    return np.random.multivariate_normal(mean, cov, N)


def circle_generate_sample(N, noise=0.25):
    angle = np.random.uniform(high=2 * np.pi, size=N)
    random_noise = np.random.normal(scale=np.sqrt(0.2), size=(N, 2))
    pos = np.concatenate([np.cos(angle), np.sin(angle)])
    pos = rearrange(pos, "(b c) -> c b", b=2)
    return pos + noise * random_noise


def olympic_generate_sample(N, noise=0.25):
    w = 3.5
    h = 1.5
    centers = np.array([[-w, h], [0.0, h], [w, h], [-w * 0.6, -h], [w * 0.6, -h]])
    pos = [
        circle_generate_sample(N // 5, noise) + centers[i : i + 1] / 2 for i in range(5)
    ]
    return np.concatenate(pos)


def four_generate_sample(N, noise=0.25):
    w = 3.5
    h = 1.5
    centers = np.array([[0.0, h], [w, h], [-w * 0.6, -h], [w * 0.6, -h]])
    pos = [
        circle_generate_sample(N // 4, noise) + centers[i : i + 1] / 2 for i in range(4)
    ]
    return np.concatenate(pos)


def spirals_sample(N, noise=0.25):
    n = np.sqrt(np.random.rand(N // 2, 1)) * 540 * (2 * np.pi) / 360
    d1x = -np.cos(n) * n + np.random.rand(N // 2, 1) * 0.5
    d1y = np.sin(n) * n + np.random.rand(N // 2, 1) * 0.5
    x = np.vstack((np.hstack((d1x, d1y)), np.hstack((-d1x, -d1y)))) / 3
    x += np.random.randn(*x.shape) * 0.1
    return x


def gaussian_sample(N, noise=0.25):
    scale = 4.0
    centers = [
        (1, 0),
        (-1, 0),
        (0, 1),
        (0, -1),
        (1.0 / np.sqrt(2), 1.0 / np.sqrt(2)),
        (1.0 / np.sqrt(2), -1.0 / np.sqrt(2)),
        (-1.0 / np.sqrt(2), 1.0 / np.sqrt(2)),
        (-1.0 / np.sqrt(2), -1.0 / np.sqrt(2)),
    ]
    centers = [(scale * x, scale * y) for x, y in centers]

    dataset = []
    for _ in range(N):
        point = np.random.randn(2) * 0.5
        idx = np.random.randint(8)
        center = centers[idx]
        point[0] += center[0]
        point[1] += center[1]
        dataset.append(point)
    dataset = np.array(dataset, dtype="float32")
    dataset /= 1.414
    return dataset


SKD_FUNC = {
    "swissroll": swissroll_generate_sample,
    "checkerboard": checkerboard_generate_sample,
    "line": line_generate_sample,
    "circle": circle_generate_sample,
    "olympic": olympic_generate_sample,
    "moon": moon_generate_sample,
    "four": four_generate_sample,
    "8gaussian": gaussian_sample,
    "2spirals": spirals_sample,
}


class Points2Dataset(Dataset):
    def __init__(
        self, data_name, num_sample, noise, dim_range=1.0, iscenter=True, shuffle=True
    ):
        self.name = data_name
        self.num_sample = num_sample
        self.noise = noise
        self.iscenter = iscenter
        self.dim_linspace = np.linspace(-dim_range, dim_range, 200)
        self.dim_range = dim_range
        self.shuffle = shuffle

        self.data = self.generate_sample()

    def generate_sample(self):
        data = SKD_FUNC[self.name](self.num_sample, self.noise)
        _max = np.max(np.abs(data))
        data = (self.dim_range * 0.85 / _max) * data
        if self.shuffle:
            np.random.shuffle(data)
        return data

    def __len__(self):
        return self.num_sample - 1

    def __getitem__(self, idx):
        return self.data[idx]



from jammy.utils.notifier import jam_notifier

if __name__ == "__main__":
    jam_notifier.notify("sending from jam", "JAM_NOTIFIER Test")
    jam_notifier.notify("second send from jam", "JAM_NOTIFIER Test")



import os.path as osp

import torch

__all__ = [
    "LossException",
    "NanException",
    "InfException",
    "step_lr",
    "register_grad_clip",
    "trainer_save_cfg",
    "check_loss_error",
]

# pylint: disable=unused-argument
class LossException(Exception):
    pass


class NanException(LossException):
    pass


class InfException(LossException):
    pass


def step_lr(trainer, *args, **kwargs):
    if trainer.lr_scheduler:
        trainer.lr_scheduler.step()
        trainer.cur_monitor.update(
            {
                "lr": trainer.optimizer.param_groups[0]["lr"],
            }
        )


def register_grad_clip(trainer, grad_clip_value):
    def grad_clip_fn(trainer, *args, **kwargs):
        torch.nn.utils.clip_grad_norm_(trainer.model.parameters(), grad_clip_value,error_if_nonfinite=False)

    trainer.register_event("backward:after", grad_clip_fn, False)


def trainer_save_cfg(trainer, cfg):
    def save_cfg(_trainer, state_dict):
        state_dict["cfg"] = cfg

    trainer.register_event("trainer:export", save_cfg, False)


def check_loss_error(trainer):
    def _deal_error(_trainer):
        model = (
            _trainer.model.module
            if hasattr(_trainer.model, "module")
            else _trainer.model
        )
        from jammy.io import locate_newest_file
        from jammy.logging import get_logger
        from jamtorch.io import load_ckpt

        fckpt = locate_newest_file(_trainer.ckpt_dir, "*.pth")
        state = load_ckpt(_trainer.device, osp.join(_trainer.ckpt_dir, fckpt))
        msg_model = model.load_state_dict(state["model"])
        logger = get_logger()
        trainer_id = trainer.rank if hasattr(trainer, "rank") else 0
        logger.critical(f"{trainer_id}: reload model ckpt {msg_model}")

    def _check_loss_error(_trainer, batch, loss, *args):
        if torch.isnan(loss):
            _deal_error(_trainer)
            raise NanException
        if torch.isinf(loss):
            _deal_error(_trainer)
            raise InfException

    trainer.register_event("forward:after", _check_loss_error, False)
    # trainer.register_event("val:step", _check_loss_error, False)



# pylint: disable=redefined-builtin,arguments-differ
from io import BytesIO
from typing import IO, Any

import numpy as np

from .base import BaseFileHandler


class NumpyHandler(BaseFileHandler):
    str_like = False

    def load_from_fileobj(self, file: IO[bytes], **kwargs) -> Any:
        """
        Load a NumPy array from a file-like object.

        Parameters:
            file (IO[bytes]): The file-like object containing the NumPy array data.
            **kwargs: Additional keyword arguments passed to `np.load`.

        Returns:
            numpy.ndarray: The loaded NumPy array.
        """
        return np.load(file, **kwargs)

    def load_from_path(self, filepath: str, **kwargs) -> Any:
        """
        Load a NumPy array from a file path.

        Parameters:
            filepath (str): The path to the file to load.
            **kwargs: Additional keyword arguments passed to `np.load`.

        Returns:
            numpy.ndarray: The loaded NumPy array.
        """
        return super().load_from_path(filepath, mode="rb", **kwargs)

    def dump_to_str(self, obj: np.ndarray, **kwargs) -> str:
        """
        Serialize a NumPy array to a string in binary format.

        Parameters:
            obj (np.ndarray): The NumPy array to serialize.
            **kwargs: Additional keyword arguments passed to `np.save`.

        Returns:
            str: The serialized NumPy array as a string.
        """
        with BytesIO() as f:
            np.save(f, obj, **kwargs)
            return f.getvalue()

    def dump_to_fileobj(self, obj: np.ndarray, file: IO[bytes], **kwargs):
        """
        Dump a NumPy array to a file-like object.

        Parameters:
            obj (np.ndarray): The NumPy array to dump.
            file (IO[bytes]): The file-like object to which the array is dumped.
            **kwargs: Additional keyword arguments passed to `np.save`.
        """
        np.save(file, obj, **kwargs)

    def dump_to_path(self, obj: np.ndarray, filepath: str, **kwargs):
        """
        Dump a NumPy array to a file path.

        Parameters:
            obj (np.ndarray): The NumPy array to dump.
            filepath (str): The file path where the array should be saved.
            **kwargs: Additional keyword arguments passed to `np.save`.
        """
        with open(filepath, "wb") as f:
            np.save(f, obj, **kwargs)



# pylint: disable=too-many-arguments, import-outside-toplevel, import-error
import io
import json
import os

from jammy.logging import get_logger
from jammy.storage.kv.memcached import MemcachedKVStore

logger = get_logger()


class Boto3Client:
    def __init__(
        self,
        conf_path: str,
        max_attempt: int = 3,
        enable_mc=False,
        mc_addr=None,
        mc_port=None,
    ):
        self.max_attempt = max_attempt
        try:
            import boto3
        except ImportError:
            raise ImportError("Please install boto3 to use Boto3Client.")

        with open(conf_path, "r") as f:
            conf = json.load(f)
        self._client = boto3.client("s3", **conf)
        if enable_mc:
            try:
                import memcache
            except ImportError:
                raise ImportError(
                    "Please install python-memcached to use Boto3Client with memcached."
                )
            self._mc_kv_store = MemcachedKVStore(mc_addr, mc_port)
        else:
            self._mc_kv_store = None

    def get(self, filepath):
        filepath = self._check_path(filepath)

        if self._mc_kv_store and self._mc_kv_store.available:
            if self._mc_kv_store.has(filepath):
                return self._mc_kv_store.get(filepath)

        attempt = 0
        while attempt < self.max_attempt:
            try:
                buffer = io.BytesIO()
                self._client.download_fileobj(
                    Bucket=filepath.split("/")[0],
                    Key="/".join(filepath.split("/")[1:]),
                    Fileobj=buffer,
                )
                buffer.seek(0)
                if self._mc_kv_store and self._mc_kv_store.available:
                    self._mc_kv_store.put(filepath, buffer.read())

                return buffer.read()
            except Exception as e:
                attempt += 1
                logger.info(f"Got an exception: attempt={attempt} - {e} - {filepath}")

        raise ConnectionError(
            "Unable to read {} from. {} attempts tried.".format(filepath, attempt)
        )

    def put(self, obj, filepath):
        filepath = self._check_path(filepath)
        bucket_name = filepath.split("/")[0]
        key = "/".join(filepath.split("/")[1:])
        from botocore.exceptions import ClientError

        attempt = 0
        while attempt < self.max_attempt:
            try:
                # If obj is a string path to a local file, use upload_file instead
                if isinstance(obj, str) and os.path.isfile(obj):
                    self._client.upload_file(Filename=obj, Bucket=bucket_name, Key=key)
                    return
                if isinstance(obj, io.BytesIO):
                    obj.seek(0)
                    self._client.upload_fileobj(obj, Bucket=bucket_name, Key=key)
                    return
                if isinstance(obj, bytes):
                    self._client.put_object(Body=obj, Bucket=bucket_name, Key=key)
                    return
                else:
                    raise ValueError("Unsupported object type for upload")
            except ClientError as e:
                attempt += 1
                logger.info(f"Got an exception: attempt={attempt} - {e} - {filepath}")

        raise ConnectionError(
            "Unable to write {} to. {} attempts tried.".format(filepath, attempt)
        )

    def contains(self, filepath: str):
        filepath = self._check_path(filepath)
        from botocore.exceptions import ClientError

        try:
            self._client.head_object(
                Bucket=filepath.split("/")[0], Key="/".join(filepath.split("/")[1:])
            )
            return True
        except ClientError:
            return False

    def isdir(self, filepath: str):
        if self.contains(filepath):
            filepath = self._check_path(filepath)
            if not filepath.endswith("/"):
                filepath += "/"
            resp = self._client.list_objects_v2(
                Bucket=filepath.split("/")[0],
                Prefix="/".join(filepath.split("/")[1:]),
                Delimiter="/",
                MaxKeys=1,
            )
            return "Contents" in resp
        return False

    def delete(self, filepath):
        filepath = self._check_path(filepath)
        self._client.delete_object(
            Bucket=filepath.split("/")[0], Key="/".join(filepath.split("/")[1:])
        )

    def list(self, filepath):
        filepath = self._check_path(filepath)
        resp = self._client.list_objects_v2(
            Bucket=filepath.split("/")[0], Prefix="/".join(filepath.split("/")[1:])
        )
        return [f["Key"] for f in resp["Contents"]]

    def _check_path(self, filepath: str):
        assert filepath.startswith("s3://")
        filepath = filepath[5:]
        return filepath



from typing import Container, Optional

import torch.distributed as dist
from torch import nn

if dist.is_available():
    from torch.distributed.distributed_c10d import _get_default_group
    from torch.distributed.utils import (
        _sync_module_states,
        _verify_param_shape_across_processes,
    )

__all__ = [
    "sync_model_states",
]


def sync_model_states(
    model: nn.Module,
    process_group: Optional[dist.ProcessGroup] = None,
    src: int = 0,
    params_and_buffers_to_ignore: Optional[Container[str]] = None,
    broadcast_buffers: bool = True,
):
    """
    Synchronizes the parameters and buffers of a model across different processes in a distributed setting.

    This function ensures that all processes in the specified process group have the same initial parameters and
    buffers from the source rank, typically rank 0. It is useful when different processes start with different model
    states and a synchronization is required to ensure consistency across all ranks.

    Args:
        model (nn.Module): The model whose parameters and buffers are to be synchronized.
        process_group (dist.ProcessGroup, optional): The process group for communication. If None,
            the default group is used. Defaults to None.
        src (int, optional): The source rank from which parameters and buffers will be broadcasted.
            Defaults to 0.
        params_and_buffers_to_ignore (Optional[Container[str]], optional): A container of parameter and buffer
            names to exclude from synchronization. Defaults to None, which means all parameters and buffers are
            included.
        broadcast_buffers (bool, optional): Whether to broadcast buffers or not. Defaults to True.

    Side Effects:
        This function modifies the state of the model in-place to synchronize it with the source rank's model state.

    Raises:
        RuntimeError: If the shapes of parameters across processes do not match, a runtime error will be raised.
    """
    if process_group is None:
        process_group = _get_default_group()
    if not params_and_buffers_to_ignore:
        params_and_buffers_to_ignore = set()

    # Build tuple of (module, parameter) for all parameters that require grads.
    modules_and_parameters = [
        (module, parameter)
        for module_name, module in model.named_modules()
        for parameter in [
            param
            # Note that we access module.named_parameters instead of
            # parameters(module). parameters(module) is only needed in the
            # single-process multi device case, where it accesses replicated
            # parameters through _former_parameters.
            for param_name, param in module.named_parameters(recurse=False)
            if param.requires_grad
            and f"{module_name}.{param_name}" not in params_and_buffers_to_ignore
        ]
    ]

    # Deduplicate any parameters that might be shared across child modules.
    memo = set()
    modules_and_parameters = [
        # "p not in memo" is the deduplication check.
        # "not memo.add(p)" is always True, and it's only there to cause "add(p)" if needed.
        (m, p)
        for m, p in modules_and_parameters
        if p not in memo and not memo.add(p)  # type: ignore[func-returns-value]
    ]

    # Build list of parameters.
    parameters = [parameter for _, parameter in modules_and_parameters]

    _verify_param_shape_across_processes(process_group, parameters)

    _sync_module_states(
        module=model,
        process_group=process_group,
        broadcast_bucket_size=int(250 * 1024 * 1024),
        src=src,
        params_and_buffers_to_ignore=params_and_buffers_to_ignore,
        broadcast_buffers=broadcast_buffers,
    )



from torch import nn

__all__ = ["MLP"]


class MLP(nn.Module):
    def __init__(self, in_dim, n_hidden, h_dim, out_dim, is_batchnorm=False):
        super().__init__()
        layers = []
        for _ in range(n_hidden):
            layers.append(nn.Linear(in_dim, h_dim))
            if is_batchnorm:
                layers.append(nn.BatchNorm1d(h_dim))
            layers.append(nn.GELU())
            in_dim = h_dim
        layers.append(nn.Linear(h_dim, out_dim))
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)



import re
from typing import Dict

import numpy as np
import pandas as pd

try:
    import plotly.express as px
except ImportError:
    px = None
import torch


def log_weight_plot(  # pylint: disable=too-many-locals
    model_cur_sd: Dict[str, torch.Tensor],
    model_prev_sd: Dict[str, torch.Tensor],
    step: int,
    html_fp="/tmp/weight_distribution_changes.html",
) -> str:
    """
    Generates and logs a plot of the weight distribution and changes between two model states.

    Args:
        model_cur_sd (Dict[str, torch.Tensor]): The current state dictionary of the model,
            containing parameter names and their corresponding tensors.
        model_prev_sd (Dict[str, torch.Tensor]): The previous state dictionary of the model,
            containing parameter names and their corresponding tensors.
        step (int): The current training step, used for labeling the plot.
        html_fp (str): The file path to save the HTML plot. Defaults to "/tmp/weight_distribution_changes.html".

    Returns:
        str: The path to the saved HTML plot.
    """
    # Initialize lists to store plot data
    layer_names = []
    std_devs = []
    l1_norms = []
    param_counts = []
    colors = []
    markers = []

    # Iterate over the parameters and compute necessary metrics
    for name, param in model_cur_sd.items():
        if name in model_prev_sd:
            prev_param = model_prev_sd[name]
            std_dev = param.std().item()
            l1_norm = torch.abs(param - prev_param).mean().item()
            param_count = param.numel()

            # Determine color based on the layer number using regex
            layer_match = re.match(r".*\.h\.(\d+)(?:\..*)?$", name)
            if layer_match:
                layer_num = int(layer_match.group(1))
                colors.append(layer_num)
            else:
                colors.append(-1)  # Non-layer parameters

            # Determine marker type based on the parameter's dimensionality
            if param.ndim == 1:
                markers.append("x")
            else:
                markers.append("circle")

            # Append data to the respective lists
            layer_names.append(name)
            std_devs.append(std_dev)
            l1_norms.append(np.log1p(l1_norm))  # log(1 + x) transformation
            param_counts.append(np.log(param_count))

    # Create a DataFrame for the plot
    df = pd.DataFrame(
        {
            "Layer Name": layer_names,
            "Standard Deviation": std_devs,
            "L1 Norm of Changes (log scale)": l1_norms,
            "Parameter Count (log)": param_counts,
            "Color": colors,
            "Marker": markers,
        }
    )

    # Determine the number of layers for color mapping
    max_layer_num = df[df["Color"] != -1]["Color"].max()

    # Create a color scale for the layers (yellow to red)
    color_scale = px.colors.sequential.YlOrRd
    color_discrete_map = {
        i: color_scale[int(i * (len(color_scale) - 1) / max_layer_num)]
        for i in range(int(max_layer_num) + 1)
    }
    color_discrete_map[-1] = "blue"  # Blue for non-layer parameters

    # Create Plotly figure
    fig = px.scatter(
        df,
        x="Standard Deviation",
        y="L1 Norm of Changes (log scale)",
        size="Parameter Count (log)",
        color="Color",
        hover_name="Layer Name",
        title=f"Model Weight Distribution and Changes, Step {step}",
        symbol="Marker",
        color_discrete_map=color_discrete_map,
        opacity=0.7,
    )

    # Save the plot as an HTML file
    fig.write_html(html_fp, auto_play=False)

    return html_fp



# Consider thread safe and singleton pattern
import inspect
from collections import defaultdict

__all__ = ["BufferCnt", "CBCnt", "ExitCnt", "bufcnt"]

g_caller_cnt = defaultdict(lambda: 0)


def bufcnt(cond):
    frames = inspect.stack()
    hash_info = "".join(
        [f"{cur_st.filename}{cur_st.lineno}\n" for cur_st in frames[1:]]
    )
    g_caller_cnt[hash_info] = g_caller_cnt[hash_info] * cond + cond
    return g_caller_cnt[hash_info]


class BufferCnt:
    def __init__(self, thres=float("inf"), reset_over_thres=False):
        self._cnt = 0
        self.thres = thres
        self.reset_over_thres = reset_over_thres

    def __call__(self, expre, thres=None):
        if expre is True:
            self._cnt += 1
        else:
            self._cnt = 0

        if thres is None:
            thres = self.thres

        if self._cnt > self.thres:
            if self.reset_over_thres:
                self.reset()
            return True

        return False

    @property
    def cnt(self):
        return self._cnt

    def reset(self):
        self._cnt = 0


class CBCnt(BufferCnt):
    def __init__(self, cb_fn=None, thres=float("inf"), reset_over_thres=False):
        super().__init__(thres, reset_over_thres=reset_over_thres)
        self.call_back = cb_fn

    def __call__(self, expre, thres=None, *args, **kwargs):
        is_thres = super().__call__(expre, thres)
        if is_thres and self.call_back is not None:
            self.call_back(*args, **kwargs)


class ExitCnt(CBCnt):
    def __init__(
        self, msg="Something wrong", thres=float("inf"), reset_over_thres=False
    ):
        def _exit_fn():
            print(msg)
            import sys

            sys.exit()

        super().__init__(_exit_fn, thres, reset_over_thres)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : sc_client.py
# Author : Jiayuan Mao, Qsh.Zh
# Email  : qsh.zh27@gmail.com
# Date   : 11/17/2021
#
# Qinsheng modifies based on Jacinle.
# Distributed under terms of the MIT license.

import sys
import time
import uuid

from jammy.comm.cs import ClientPipe


def main():
    client = ClientPipe("client" + uuid.uuid4().hex[:8], conn_info=sys.argv[1:3])
    print("Identity: {}.".format(client.identity))
    with client.activate():
        in_ = dict(a=1, b=2)
        out = client.query("calc", in_)
        print("Success: input={}, output={}".format(in_, out))
        time.sleep(1)


if __name__ == "__main__":
    main()



import functools

import numpy as np
import six
import torch

from jammy.utils.meta import stmap

SKIP_TYPES = six.string_types

__all__ = [
    "as_tensor",
    "as_numpy",
    "as_float",
    "as_cuda",
    "as_device",
    "as_cpu",
    "as_detached",
]


def _as_tensor(cur_obj):
    if isinstance(cur_obj, SKIP_TYPES):
        return cur_obj
    if torch.is_tensor(cur_obj):
        return cur_obj
    return torch.from_numpy(np.array(cur_obj))


def as_tensor(obj):
    return stmap(_as_tensor, obj)


def _as_numpy(cur_obj):
    if isinstance(cur_obj, SKIP_TYPES):
        return cur_obj
    if torch.is_tensor(cur_obj):
        return cur_obj.cpu().numpy()
    return np.array(cur_obj)


def as_numpy(obj):
    return stmap(_as_numpy, obj)


def _as_float(cur_obj):
    if isinstance(cur_obj, SKIP_TYPES):
        return cur_obj
    if torch.is_tensor(cur_obj):
        return cur_obj.item()
    arr = as_numpy(cur_obj)
    assert arr.size == 1
    return float(arr)


def as_float(obj):
    return stmap(_as_float, obj)


def _as_cuda(cur_obj):
    if torch.is_tensor(cur_obj):
        return cur_obj.cuda()
    return cur_obj


def as_cuda(obj):
    return stmap(_as_cuda, obj)


def _as_device(cur_obj, device):
    if isinstance(cur_obj, SKIP_TYPES):
        return cur_obj
    if torch.is_tensor(cur_obj):
        cur_obj.to(device)
    return cur_obj


def as_device(obj, device=torch.device("cuda:0")):
    return stmap(functools.partial(_as_device, device=device), obj)


as_cpu = functools.partial(as_device, device=torch.device("cpu"))


def _as_detached(cur_obj, clone=False):
    if torch.is_tensor(cur_obj):
        if clone:
            return cur_obj.clone().detach()
        return cur_obj.detach()
    return cur_obj


def as_detached(obj, clone=False):
    return stmap(functools.partial(_as_detached, clone=clone), obj)



import tempfile as tempfile_lib
import contextlib
import os

__all__ = ["tempfile"]


@contextlib.contextmanager
def tempfile(mode="w+b", suffix="", prefix="tmp"):
    f = tempfile_lib.NamedTemporaryFile(
        mode, suffix=suffix, prefix=prefix, delete=False
    )
    yield f
    os.unlink(f.name)



# pylint: disable=invalid-name
import os
import os.path as osp
import sys
import tempfile

import yaml

from jammy.logging import get_logger

# logger = get_logger(sys.stdout, colorize=True,
#                   format="<green>{time}</green> <level>{message}</level>")
logger = get_logger(sys.stderr, colorize=True, level="INFO")


def load_system_settings(root, config, bash_file):  # pylint: disable=unused-argument
    if "system" not in config or config["system"] is None:
        return
    envs = config["system"].get("envs", {})
    for k, v in envs.items():
        logger.info("Export system environment variable {} = {}.".format(k, v))
        print("export {}={}".format(k, v), file=bash_file)


def load_vendors(root, config, bash_file):  # pylint: disable=unused-argument
    if "vendors" not in config or config["vendors"] is None:
        return

    for k, v in config["vendors"].items():
        assert "root" in v, '"root" not found in vendor: {}.'.format(k)

        logger.info("Loading vendor: {}.".format(k))
        print(
            "export PYTHONPATH={}:$PYTHONPATH".format(osp.join(root, v["root"])),
            file=bash_file,
        )


def load_conda_settings(root, config, bash_file):  # pylint: disable=unused-argument
    if "conda" not in config or config["conda"] is None:
        return
    target_env = config["conda"].get("env", "")
    if target_env != "":
        logger.info("Using conda env: {}.".format(target_env))
        print(
            """
if [[ $CONDA_DEFAULT_ENV != "{v}" ]]; then
    eval "$(conda shell.bash hook)"
    conda activate {v}
fi
""".format(
                v=target_env
            ),
            file=bash_file,
        )


def load_yml_config(root, bash_file, recursive=False):
    if recursive:
        last_root = None
        while root != last_root:
            yml_filename = osp.join(root, "jam_.yaml")
            if osp.isfile(yml_filename):
                break
            last_root = root
            root = osp.dirname(root)
    else:
        yml_filename = osp.join(root, "jam_.yaml")

    if osp.isfile(yml_filename):
        logger.critical("Loading jam config: {}.".format(osp.abspath(yml_filename)))
        with open(yml_filename) as f:
            config = yaml.safe_load(f.read())
        if config is not None:
            load_system_settings(root, config, bash_file)
            load_vendors(root, config, bash_file)
            load_conda_settings(root, config, bash_file)


def main():
    f = tempfile.NamedTemporaryFile("w", delete=False)
    load_yml_config(
        osp.dirname(osp.dirname(__file__)), f
    )  # jammy package configuration
    load_yml_config(os.getcwd(), f, recursive=True)  # project configuration
    f.close()
    print(f.name)


if __name__ == "__main__":
    main()



from jammy.utils.py_types import is_list_of

from .base import BaseFileHandler
from .imageio_video_handler import ImageioVideoHandler
from .json_handler import JsonHandler
from .np_handler import NumpyHandler
from .pickle_handler import PickleHandler
from .pil_handler import PILHandler
from .tarfile_handler import TarHandler
from .torch_handler import TorchHandler
from .torchjit_handler import TorchJitHandler
from .yaml_handler import YamlHandler

file_handlers = {
    "json": JsonHandler(),
    "yaml": YamlHandler(),
    "yml": YamlHandler(),
    "pickle": PickleHandler(),
    "pkl": PickleHandler(),
    "tar": TarHandler(),
    "jit": TorchJitHandler(),
    "npy": NumpyHandler(),
}

for torch_type in ["pt", "pth", "ckpt"]:
    file_handlers[torch_type] = TorchHandler()
for img_type in ["jpg", "jpeg", "png", "bmp", "gif"]:
    file_handlers[img_type] = PILHandler()
for video_type in ["mp4", "avi", "mov", "webm", "flv", "wmv"]:
    file_handlers[video_type] = ImageioVideoHandler()


def _register_handler(handler, file_formats):
    """Register a handler for some file extensions.

    Args:
        handler (:obj:`BaseFileHandler`): Handler to be registered.
        file_formats (str or list[str]): File formats to be handled by this
            handler.
    """
    if not isinstance(handler, BaseFileHandler):
        raise TypeError(
            f"handler must be a child of BaseFileHandler, not {type(handler)}"
        )
    if isinstance(file_formats, str):
        file_formats = [file_formats]
    if not is_list_of(file_formats, str):
        raise TypeError("file_formats must be a str or a list of str")
    for ext in file_formats:
        file_handlers[ext] = handler


def register_handler(file_formats, **kwargs):
    def wrap(cls):
        _register_handler(cls(**kwargs), file_formats)
        return cls

    return wrap



from jammy.utils.env import jam_getenv

try:
    from .utils import *
except ImportError:
    from jammy.logging import get_logger

    logger = get_logger()
    logger.critical("Import Error, check dependency of torch packages")


from jamtorch.utils.init import init_main

init_main()

del init_main

if jam_getenv("IMPORT_ALL", "true", "bool"):
    from jamtorch.cuda import set_best_device
    from jamtorch.logging import get_logger  # pylint: disable=ungrouped-imports



from typing import Any

from jammy import logging

from ..distributed import get_rank, is_rank0

try:
    import torch_xla.core.xla_model as xm  # pylint: disable=import-error

    IS_TPU = True
except Exception as e:  # pylint: disable=broad-except
    IS_TPU = False

__all__ = ["get_logger"]


def get_logger(*args, **kwargs):
    if IS_TPU:
        if xm.get_ordinal == 0:
            return logging.get_logger(*args, **kwargs)
    elif is_rank0():
        return logging.get_logger(*args, **kwargs)

    return logging.fake_logger


def loguru_rank0_only_filter(record: Any) -> bool:
    rank0_only = record["extra"].get("rank0_only", True)
    if is_rank0() and rank0_only:
        return True
    if not rank0_only:
        record["message"] = f"[RANK {get_rank()}]" + record["message"]
    return not is_rank0



try:
    import torch
except ImportError:
    torch = None

from .base import BaseFileHandler


class TorchHandler(BaseFileHandler):

    str_like = False

    def load_from_fileobj(self, file, **kwargs):
        return torch.load(file, **kwargs)

    def dump_to_fileobj(self, obj, file, **kwargs):
        torch.save(obj, file, **kwargs)

    def dump_to_str(self, obj, **kwargs):
        raise NotImplementedError



__all__ = ["fake_logger"]

class FakeLogger:
    def info(*args, **kwargs):
        pass

    def debug(*args, **kwargs):
        pass

    def trace(*args, **kwargs):
        pass

    def parse(*args, **kwargs):
        pass

    def warning(*args, **kwargs):
        pass

    def critical(*args, **kwargs):
        pass

    def log(*args, **kwargs):
        pass

    def error(*args, **kwargs):
        pass

    def success(*args, **kwargs):
        pass

fake_logger = FakeLogger()


# Copyright (c) OpenMMLab. All rights reserved.
from .backends import (
    BaseStorageBackend,
    Boto3Backend,
    HTTPBackend,
    LmdbBackend,
    LocalBackend,
    MemcachedBackend,
    register_backend,
)
from .file_client import FileClient, HardDiskBackend
from .handlers import (
    BaseFileHandler,
    JsonHandler,
    PickleHandler,
    YamlHandler,
    register_handler,
)
from .io import (
    copy_if_symlink_fails,
    copyfile,
    copyfile_from_local,
    copyfile_to_local,
    copytree,
    copytree_from_local,
    copytree_to_local,
    dump,
    exists,
    generate_presigned_url,
    get,
    get_file_backend,
    get_local_path,
    get_text,
    isdir,
    isfile,
    join_path,
    list_dir_or_file,
    load,
    put,
    put_text,
    remove,
    rmtree,
    set_s3_backend,
)
from .parse import dict_from_file, list_from_file

__all__ = [
    "BaseStorageBackend",
    "FileClient",
    "Boto3Backend",
    "MemcachedBackend",
    "LmdbBackend",
    "HardDiskBackend",
    "LocalBackend",
    "HTTPBackend",
    "copy_if_symlink_fails",
    "copyfile",
    "copyfile_from_local",
    "copyfile_to_local",
    "copytree",
    "copytree_from_local",
    "copytree_to_local",
    "exists",
    "generate_presigned_url",
    "get",
    "get_file_backend",
    "get_local_path",
    "get_text",
    "isdir",
    "isfile",
    "join_path",
    "list_dir_or_file",
    "put",
    "put_text",
    "remove",
    "rmtree",
    "load",
    "dump",
    "register_handler",
    "BaseFileHandler",
    "JsonHandler",
    "PickleHandler",
    "YamlHandler",
    "list_from_file",
    "dict_from_file",
    "register_backend",
    "set_s3_backend",
]



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D

from .color_list import jcolors

__all__ = ["MstdDict", "mstd_plot", "plotstd"]

color_list = jcolors


def MstdDict():
    return defaultdict(lambda: defaultdict(list))


# pylint: disable=too-many-locals
def point_stat(data, coef_std=1, **kwargs):
    mean = np.mean(data)
    if "ylog" in kwargs:
        std = np.std(data)
        above = np.log(mean + std)
        down = np.log(mean - std)
        space = coef_std * np.min([above - np.log(mean), np.log(mean) - down])
        return {
            "mean": mean,
            "up": np.exp(np.log(mean) + space),
            "down": np.exp(np.log(mean) - space),
        }
    else:
        return {
            "mean": mean,
            "up": mean + coef_std * np.std(data),
            "down": mean - coef_std * np.std(data),
        }


def plotstd(mean, cov, x=None, color=None, ax=None, label=None):
    if ax is None:
        _, ax = plt.subplots(1, 1)
    if color is None:
        color = color_list[0]

    mean = np.array(mean).flatten()
    cov = np.array(cov).flatten()
    if x is None:
        x = np.arange(len(mean))

    ax.plot(x, mean, "o", color=color)
    ax.plot(x, mean, "-", color=color, label=label)
    ax.fill_between(x, mean - cov, mean + cov, color=color, alpha=0.2)

    return ax


def mstd_plot(  # pylint: disable=too-many-branches
    exp_data,
    labels=None,
    size=(7, 7),
    ax=None,
    is_label=True,
    coef_std=1,
    marker=None,
    **kwargs,
):
    """plot mean and std of sequence

    :param exp_data: loaded data, key:method->key:x_value->list of y
    :type exp_data: dict
    :param labels: key:method->value:figure labels, or None
    :type labels: dict
    :param size: figure_size, defaults to (7, 7)
    :param ax: maptlotlib axis, defaults to None
    :param is_label: whether or not show label, defaults to True
    :param coef_std: width of std, defaults to 1
    :return: ax

    Example::
        exp_data = {
            "method_A": {
                            1: [seed1_y, seed2_y, seed3_y],
                            2: [seed1_y, seed2_y, seed3_y],
                        },
            "method_B": {
                            1: [seed1_y, seed2_y, seed3_y],
                            2: [seed1_y, seed2_y, seed3_y]
                        }
        }
        labels = {
            "method_A": "name_A_in_paper",
            "method_B": "name_B_in_paper",
        }
    """
    global color_list  # pylint: disable=global-variable-not-assigned
    if marker is None:
        marker = [
            "-",
        ]
    if labels is None:
        labels = {key_: key_ for key_ in exp_data.keys()}
    rtn_dict = {}
    for exp_name in labels.keys():
        cur = []
        for _y_list in exp_data[exp_name].values():
            cur.append(list(point_stat(_y_list, coef_std, **kwargs).values()))

        cur = np.array(cur)
        rtn_dict[exp_name] = {
            "x": np.array(list(exp_data[exp_name].keys())),
            "mean": cur[:, 0],
            "up": cur[:, 1],
            "down": cur[:, 2],
        }

    if ax is None:
        _, ax = plt.subplots(figsize=size)
    for i, exp_name in enumerate(labels):
        x = rtn_dict[exp_name]["x"]
        mean = rtn_dict[exp_name]["mean"]
        above = rtn_dict[exp_name]["up"]
        down = rtn_dict[exp_name]["down"]
        cur_color = color_list[i % len(color_list)]
        # ax.plot(x, mean, "o", color=cur_color)
        # ax.plot(x, mean, "-", color=cur_color)
        for cur_marker in marker:
            ax.plot(x, mean, cur_marker, color=cur_color)
        ax.fill_between(x, down, above, color=color_list[i], alpha=0.2)

    if "fix_legend" in kwargs:
        legend_label = [f"{kwargs['fix_legend']}={value}" for value in labels.values()]
    else:
        legend_label = list(labels.values())

    if is_label:
        custom_lines = [
            Line2D([0], [0], color=color_list[i], lw=4) for i, _ in enumerate(labels)
        ]
        #        ax.legend(custom_lines, list(labels.values()))
        ax.legend(custom_lines, legend_label)

    if "xlabel" in kwargs:
        ax.set_xlabel(kwargs["xlabel"])
    if "ylabel" in kwargs:
        ax.set_ylabel(kwargs["ylabel"])
    if "ylog" in kwargs:
        ax.set_yscale("log")

    fig = plt.gcf()

    return fig, ax



from .argument import *
from .cache import *
from .cfg import *
from .cnt import *
from .hash import *
from .meta import Singleton
from .profiler import IS_PROFILING, profile, profile_now
from .py_types import *



# Copyright (c) OpenMMLab. All rights reserved.
from pathlib import Path
from typing import Union

from .base import BaseStorageBackend


class MemcachedBackend(BaseStorageBackend):
    """Memcached storage backend.

    Attributes:
        server_list_cfg (str): Config file for memcached server list.
        client_cfg (str): Config file for memcached client.
        sys_path (str, optional): Additional path to be appended to `sys.path`.
            Defaults to None.
    """

    def __init__(self, server_list_cfg, client_cfg, sys_path=None):
        if sys_path is not None:
            import sys

            sys.path.append(sys_path)
        try:
            import mc
        except ImportError:
            raise ImportError("Please install memcached to enable MemcachedBackend.")

        self.server_list_cfg = server_list_cfg
        self.client_cfg = client_cfg
        self._client = mc.MemcachedClient.GetInstance(
            self.server_list_cfg, self.client_cfg
        )
        # mc.pyvector servers as a point which points to a memory cache
        self._mc_buffer = mc.pyvector()

    def get(self, filepath: Union[str, Path]):
        """Get values according to the filepath.

        Args:
            filepath (str or Path): Path to read data.

        Returns:
            bytes: Expected bytes object.

        Examples:
            >>> server_list_cfg = '/path/of/server_list.conf'
            >>> client_cfg = '/path/of/mc.conf'
            >>> backend = MemcachedBackend(server_list_cfg, client_cfg)
            >>> backend.get('/path/of/file')
            b'hello world'
        """
        filepath = str(filepath)
        import mc

        self._client.Get(filepath, self._mc_buffer)
        value_buf = mc.ConvertBuffer(self._mc_buffer)
        return value_buf

    def get_text(self, filepath, encoding=None):
        raise NotImplementedError



from jamtorch.jdatasets import Points2Dataset


def test_toy_dataset():
    dataset = Points2Dataset("swissroll", 1000, 0.0)
    for i in range(10):
        print(dataset[i])


# pytest -s tests/jamtorch/test_jdataset.py



import os
import os.path as osp

import git

from jammy.logging import get_logger

logger = get_logger()

__all__ = ["is_git", "git_rootdir", "git_hash"]


def is_git(path):
    try:
        _ = git.Repo(path, search_parent_directories=True).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False


def git_rootdir(path=""):
    if is_git(os.getcwd()):
        _git_repo = git.Repo(os.getcwd(), search_parent_directories=True)
        root = _git_repo.git.rev_parse("--show-toplevel")
        return osp.join(root, path)
    logger.warning("not a git repo")
    return osp.join(os.getcwd(), path)


def git_hash(path):
    if is_git(path):
        _git_repo = git.Repo(path, search_parent_directories=True)
        return _git_repo.head.object.hexsha
    logger.warning("not a git repo")
    return None


def git_repo(path):
    if is_git(path):
        _git_repo = git.Repo(path, search_parent_directories=True)
        return _git_repo
    logger.warning("not a git repo")
    return None


def log_repo(path):
    repo = git_repo(path)
    if repo:
        return repo.head.object.hexsha, repo.git.diff()
    # if not repo, return None sha, empty diff
    return None, ""
